@@ -1,6 +1,8 @@
 """Django Ninja Extra - Class Based Utility and more for Django Ninja(Fast Django REST framework)"""
 
-__version__ = "0.13.8"
+__version__ = "0.14.4"
+
+import django
 
 from ninja_extra.controllers import (
     ControllerBase,
@@ -15,8 +17,11 @@ from ninja_extra.controllers import (
 from ninja_extra.controllers.route import route
 from ninja_extra.dependency_resolver import get_injector, service_resolver
 from ninja_extra.main import NinjaExtraAPI
+from ninja_extra.router import Router
+
+if django.VERSION < (3, 2):  # pragma: no cover
+    default_app_config = "ninja_extra.apps.NinjaExtraConfig"
 
-default_app_config = "ninja_extra.apps.NinjaExtraConfig"
 
 __all__ = [
     "ControllerBase",
@@ -36,4 +41,5 @@ __all__ = [
     "get_injector",
     "service_resolver",
     "lazy",
+    "Router",
 ]
