@@ -1,3 +1,4 @@
+import inspect
 import time
 from contextlib import contextmanager
 from typing import (
@@ -8,7 +9,9 @@ from typing import (
     List,
     Optional,
     Sequence,
+    Type,
     Union,
+    cast,
 )
 
 from django.http import HttpRequest
@@ -21,8 +24,11 @@ from ninja.operation import (
     PathView as NinjaPathView,
 )
 from ninja.signature import is_async
+from ninja.types import TCallable
+from ninja.utils import check_csrf
 
 from ninja_extra.exceptions import APIException
+from ninja_extra.helper import get_function_name
 from ninja_extra.logger import request_logger
 from ninja_extra.signals import route_context_finished, route_context_started
 
@@ -35,12 +41,41 @@ if TYPE_CHECKING:
 
 class Operation(NinjaOperation):
     def __init__(
-        self, *args: Any, url_name: Optional[str] = None, **kwargs: Any
+        self,
+        path: str,
+        methods: List[str],
+        view_func: Callable,
+        *,
+        url_name: Optional[str] = None,
+        **kwargs: Any,
     ) -> None:
-        super().__init__(*args, **kwargs)
+        self.is_coroutine = is_async(view_func)
         self.url_name = url_name
+        super().__init__(path, methods, view_func, **kwargs)
         self.signature = ViewSignature(self.path, self.view_func)
 
+    def _set_auth(
+        self, auth: Optional[Union[Sequence[Callable], Callable, object]]
+    ) -> None:
+        if auth is not None and auth is not NOT_SET:
+            self.auth_callbacks = isinstance(auth, Sequence) and auth or [auth]
+            for callback in self.auth_callbacks:
+                _call_back = (
+                    callback if inspect.isfunction(callback) else callback.__call__  # type: ignore
+                )
+
+                if not getattr(callback, "is_coroutine", None):
+                    setattr(callback, "is_coroutine", is_async(_call_back))
+
+                if is_async(_call_back) and not self.is_coroutine:
+                    raise Exception(
+                        f"Could apply auth=`{get_function_name(callback)}` "
+                        f"to view_func=`{get_function_name(self.view_func)}`.\n"
+                        f"N:B - {get_function_name(callback)} can only be used on Asynchronous view functions"
+                    )
+
+
+class ControllerOperation(Operation):
     def _log_action(
         self,
         logger: Callable[..., Any],
@@ -90,10 +125,8 @@ class Operation(NinjaOperation):
             context = self.get_execution_context(request, **kw)
             # send route_context_started signal
             route_context_started.send(RouteContext, route_context=context)
-            values = self._get_values(request, kw)
-            context.kwargs = values
 
-            yield values, context
+            yield context
             self._log_action(
                 request_logger.info,
                 request=request,
@@ -115,15 +148,16 @@ class Operation(NinjaOperation):
             route_context_finished.send(RouteContext, route_context=None)
 
     def run(self, request: HttpRequest, **kw: Any) -> HttpResponseBase:
-        error = super(Operation, self)._run_checks(request)
+        error = self._run_checks(request)
         if error:
             return error
         try:
             with self._prep_run(request, **kw) as ctx:
-                values, context = ctx
-                result = self.view_func(context=context, **values)
+                values = self._get_values(request, kw)
+                ctx.kwargs = values
+                result = self.view_func(context=ctx, **values)
                 _processed_results = self._result_to_response(request, result)
-            return _processed_results
+                return _processed_results
         except Exception as e:
             if isinstance(e, TypeError) and "required positional argument" in str(e):
                 msg = "Did you fail to use functools.wraps() in a decorator?"
@@ -133,16 +167,73 @@ class Operation(NinjaOperation):
 
 
 class AsyncOperation(Operation, NinjaAsyncOperation):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
+        super().__init__(*args, **kwargs)
+        from asgiref.sync import sync_to_async
+
+        self._get_values = cast(Callable, sync_to_async(super()._get_values))  # type: ignore
+        self._result_to_response = cast(  # type: ignore
+            Callable,
+            sync_to_async(super()._result_to_response),
+        )
+
+    async def _run_checks(self, request: HttpRequest) -> Optional[HttpResponse]:  # type: ignore
+        """Runs security checks for each operation"""
+        # auth:
+        if self.auth_callbacks:
+            error = await self._run_authentication(request)
+            if error:
+                return error
+
+        # csrf:
+        if self.api.csrf:
+            error = check_csrf(request, self.view_func)
+            if error:
+                return error
+
+        return None
+
+    async def _run_authentication(self, request: HttpRequest) -> Optional[HttpResponse]:  # type: ignore
+        for callback in self.auth_callbacks:
+            try:
+                is_coroutine = getattr(callback, "is_coroutine", False)
+                if is_coroutine:
+                    result = await callback(request)
+                else:
+                    result = callback(request)
+            except Exception as exc:
+                return self.api.on_exception(request, exc)
+
+            if result:
+                request.auth = result  # type: ignore
+                return None
+        return self.api.create_response(request, {"detail": "Unauthorized"}, status=401)
+
     async def run(self, request: HttpRequest, **kw: Any) -> HttpResponseBase:  # type: ignore
-        error = self._run_checks(request)
+        error = await self._run_checks(request)
+        if error:
+            return error
+        try:
+            values = await self._get_values(request, kw)  # type: ignore
+            result = await self.view_func(request, **values)
+            _processed_results = await self._result_to_response(request, result)  # type: ignore
+            return cast(HttpResponseBase, _processed_results)
+        except Exception as e:
+            return self.api.on_exception(request, e)
+
+
+class AsyncControllerOperation(AsyncOperation, ControllerOperation):
+    async def run(self, request: HttpRequest, **kw: Any) -> HttpResponseBase:  # type: ignore
+        error = await self._run_checks(request)
         if error:
             return error
         try:
             with self._prep_run(request, **kw) as ctx:
-                values, context = ctx
-                result = await self.view_func(context=context, **values)
-                _processed_results = self._result_to_response(request, result)
-            return _processed_results
+                values = await self._get_values(request, kw)  # type: ignore
+                ctx.kwargs = values
+                result = await self.view_func(context=ctx, **values)
+                _processed_results = await self._result_to_response(request, result)  # type: ignore
+                return cast(HttpResponseBase, _processed_results)
         except Exception as e:
             return self.api.on_exception(request, e)
 
@@ -176,12 +267,7 @@ class PathView(NinjaPathView):
     ) -> Operation:
         if url_name:
             self.url_name = url_name
-
-        operation_class = Operation
-        if is_async(view_func):
-            self.is_async = True
-            operation_class = AsyncOperation
-
+        operation_class = self.get_operation_class(view_func)
         operation = operation_class(
             path,
             methods,
@@ -203,3 +289,23 @@ class PathView(NinjaPathView):
 
         self.operations.append(operation)
         return operation
+
+    def get_operation_class(
+        self, view_func: TCallable
+    ) -> Type[Union[Operation, AsyncOperation]]:
+        operation_class = Operation
+        if is_async(view_func):
+            self.is_async = True
+            operation_class = AsyncOperation
+        return operation_class
+
+
+class ControllerPathView(PathView):
+    def get_operation_class(
+        self, view_func: TCallable
+    ) -> Type[Union[Operation, AsyncOperation]]:
+        operation_class = ControllerOperation
+        if is_async(view_func):
+            self.is_async = True
+            operation_class = AsyncControllerOperation
+        return operation_class
